#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2009-04-07
# Last Modified: $Date: 2011-08-10 10:05:48 +0100 (Wed, 10 Aug 2011) $
# Id:            $Id: Fastq_Split.pm 13877 2011-08-10 09:05:48Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-38.0/lib/npg_common/Fastq_Split.pm $
#

package npg_common::Fastq_Split;

use strict;
use warnings;
use Carp;
use English qw(-no_match_vars);
use base qw(Class::Accessor);
use npg_common::Alignment;
use File::Basename;
use File::Temp qw(tempfile tempdir);
use npg_qc::autoqc::results::split_stats;
use Cwd qw(abs_path) ;

use IO::Tee;

use DB_File;
use Parallel::ForkManager;

use npg_qc::autoqc::checks::check;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13877 $ =~ /(\d+)/mxs; $r; };

__PACKAGE__->mk_accessors(qw(bwa_cmd bwa_options srf_filter_cmd samtools_cmd));

sub new {
  my ($class, $ref) = @_;

  $ref ||= {};
  bless $ref, $class;

  return $ref;
}

sub file_split{ ## no critic (Subroutines::ProhibitExcessComplexity)
  my ($self, $args) = @_;

  my $fastq1 = $args->{fastq1};
  my $fastq2 = $args->{fastq2};
  my $fastqt = $args->{fastqt};
  my $srf_filename = $args->{srf};
  my $srf_filter = $args->{srf_filter};
  my $human_ref_bwa = $args->{ref_bwa};
  my $human_ref_fa = $args->{ref_fa};
  my $output_dir = $args->{output_dir};
  my $ref_name = $args->{ref_name} || q{human};
  my $fork_alignment = $args->{fork_alignment};
  my $id_run = $args->{id_run};
  my $position = $args->{position};
  my $stats_only = $args->{stats_only};
  my $no_pileup = $args->{no_pileup};

  #
  #check necessary input
  #

  if(!$human_ref_bwa){

    croak 'Reference bwa format was not given!';
  }

  if(!$human_ref_fa){

    croak 'Reference fa format was not given!';
  }

  if(!$output_dir){

    croak 'Output directory was not given!';
  }

  if(!$fastq1 && !$fastq2){

    croak 'At least one fastq file must be provided!';
  }

  print {*STDERR} "Trying to split fastq files\n" or croak $ERRNO;

  #
  #store some basic info into result
  #
  my $split_stats_result = npg_qc::autoqc::results::split_stats->new();

  $split_stats_result->set_info( 'Check', ref $self );
  $split_stats_result->set_info('Check_version', $VERSION );

  $split_stats_result->ref_name($ref_name);
  $split_stats_result->reference(abs_path($human_ref_bwa));

  if($self->bwa_cmd()){

    my $aligner_version = npg_qc::autoqc::checks::check->aligner_version($self->bwa_cmd());

    $split_stats_result->set_info( 'Aligner', $self->bwa_cmd() );
    $split_stats_result->set_info( 'Aligner_version', $aligner_version );
  }

  if($self->bwa_options()){

    $split_stats_result->set_info( 'Aligner_options', $self->bwa_options() );
  }

  if ($id_run){
    $split_stats_result->id_run($id_run);
  }

  if($position){
    $split_stats_result->position($position);
  }

  #
  # prepare some output filename
  #
  my $fastq_list = [$fastq1, $fastq2];

  my $work_dir = tempdir(CLEANUP => 1);

  my $hit_filename1      = $work_dir.'/1.hithash';
  my $nonhit_filename1   = $work_dir.'/1.nonhithash';
  my $hit_filename2      = $work_dir.'/2.hithash';
  my $nonhit_filename2   = $work_dir.'/2.nonhithash';

  my $bam_out_dir = tempdir(CLEANUP => 1, DIR => $output_dir);

  my $bam_outputfile1    = $bam_out_dir.'/1.bam';
  my $bam_outputfile2    = $bam_out_dir.'/2.bam';


  my ($chromosome_coverage1, $chromosome_depth1, $chromosome_coverage2, $chromosome_depth2);

  my $args_ref = { work_dir         => $work_dir,
                   fastq_list       => $fastq_list,
                   human_ref_bwa    => $human_ref_bwa,
                   hit_filename1    => $hit_filename1,
                   nonhit_filename1 => $nonhit_filename1,
                   hit_filename2    => $hit_filename2,
                   nonhit_filename2 => $nonhit_filename2,
                   bam_outputfile1  => $bam_outputfile1,
                   bam_outputfile2  => $bam_outputfile2,
                 };

  #
  # doing BWA alignment here
  # generate hit and nonhit read list being tied into files and bam file for each fastq
  #
  if($fork_alignment){

    $self->bwa_alignment_list_fork( $args_ref );
  }else{

    $self->bwa_alignment_list( $args_ref );
  }

  #
  # alignement coverage and depth, and hit, nonhit hash list
  #
  my $hit_sequence_ids1;
  my $nonhit_sequence_ids1;
  my $hit_sequence_ids2;
  my $nonhit_sequence_ids2;

  if($fastq1){

    if(! $no_pileup){
       ($chromosome_coverage1, $chromosome_depth1)
       = $self->coverage_depth( $bam_outputfile1, $human_ref_fa);
    }

    unlink $bam_outputfile1;

    $split_stats_result->filename1($fastq1);
    $split_stats_result->alignment_coverage1($chromosome_coverage1);
    $split_stats_result->alignment_depth1($chromosome_depth1);

    if((! -e $hit_filename1) || (! -e $nonhit_filename1)){
       croak "hash tied file not exists: $hit_filename1 or $nonhit_filename1";
    }

    tie %{$hit_sequence_ids1}, 'DB_File', $hit_filename1 or croak "Cannot tie file $hit_filename1 to a hash\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)
    tie %{$nonhit_sequence_ids1}, 'DB_File', $nonhit_filename1 or croak "Cannot tiw file $nonhit_filename1 to a hash\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)

    $split_stats_result->num_aligned1($self->hash_size($hit_sequence_ids1));
    $split_stats_result->num_not_aligned1($self->hash_size($nonhit_sequence_ids1));

  }

  if($fastq2){
    if( !$no_pileup ){
       ($chromosome_coverage2, $chromosome_depth2)
       = $self->coverage_depth( $bam_outputfile2,  $human_ref_fa);
    }

    unlink $bam_outputfile2;

    $split_stats_result->filename2($fastq2);
    $split_stats_result->alignment_coverage2($chromosome_coverage2);
    $split_stats_result->alignment_depth2($chromosome_depth2);

    if((! -e $hit_filename2) || (! -e $nonhit_filename2)){
       croak "hash tied file not exists: $hit_filename2 or $nonhit_filename2";
    }

    tie %{$hit_sequence_ids2}, 'DB_File', $hit_filename2 or croak "Cannot tie file $hit_filename2 to a hash\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)
    tie %{$nonhit_sequence_ids2}, 'DB_File', $nonhit_filename2 or croak "Cannot open file $nonhit_filename2 to a hash\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)

    $split_stats_result->num_aligned2($self->hash_size($hit_sequence_ids2));
    $split_stats_result->num_not_aligned2($self->hash_size($nonhit_sequence_ids2));
  }

  #
  # mege hit list if two fastq file given
  if($fastq1 && $fastq2){

    if((! -e $hit_filename1) || (! -e $nonhit_filename1) || (! -e $hit_filename2) || (! -e $nonhit_filename2) ){
       croak "hash tied file not exists: $hit_filename1, $nonhit_filename1, $hit_filename2 or $nonhit_filename2";
    }
    while( my ($id, $value) = each %{$nonhit_sequence_ids1} ){
        if($hit_sequence_ids2->{$id}){
               delete $nonhit_sequence_ids1->{$id};
               $hit_sequence_ids1->{$id}++;
        }
    }
  }

  $split_stats_result->num_aligned_merge($self->hash_size($hit_sequence_ids1));
  $split_stats_result->num_not_aligned_merge($self->hash_size($nonhit_sequence_ids1));

  #
  # splitting fastq files now
  #
  print {*STDERR} "Spliting fastq files\n" or croak $ERRNO;

  my $fastq1_name;
  my $fastq2_name;

  if($fastq1){

       $fastq1_name = $self->get_filename_no_extension($fastq1);

       my $fastq1_human = $output_dir.q{/}.$fastq1_name.qq{_$ref_name.fastq};
       my $fastq1_nonhuman = $output_dir.q{/}.$fastq1_name.qq{_non$ref_name.fastq};
       if(!$stats_only){
         $self->fastq_split($fastq1, $hit_sequence_ids1, $fastq1_human, $fastq1_nonhuman);
       }
  }

  if($fastq2){

       $fastq2_name = $self->get_filename_no_extension($fastq2);
       my $fastq2_human = $output_dir.q{/}.$fastq2_name.qq{_$ref_name.fastq};
       my $fastq2_nonhuman = $output_dir.q{/}.$fastq2_name.qq{_non$ref_name.fastq};
       if(!$stats_only){
         $self->fastq_split($fastq2, $hit_sequence_ids1, $fastq2_human, $fastq2_nonhuman);
       }
  }

  if($fastqt && (-e $fastqt)){

       my $fastqt_name = $self->get_filename_no_extension($fastqt);
       my $fastqt_human = $output_dir.q{/}.$fastqt_name.qq{_$ref_name.fastq};
       my $fastqt_nonhuman = $output_dir.q{/}.$fastqt_name.qq{_non$ref_name.fastq};
       if(!$stats_only){
         $self->fastq_split($fastqt, $hit_sequence_ids1, $fastqt_human, $fastqt_nonhuman);
       }
  }

  #
  # using fastq file names to generate filter and stats file names
  #

  my $both_fastq;
  if($fastq1_name && $fastq2_name){

    $both_fastq = $fastq1_name.q{-}.$fastq2_name;
  }elsif($fastq1_name){

    $both_fastq = $fastq1_name;
  }elsif($fastq2_name){

    $both_fastq = $fastq2_name;
  }

  #
  # save stats to xml and json
  #

  my $qc_output_dir = qq{$output_dir/qc};

  if(!-e $qc_output_dir){

    mkdir $qc_output_dir;
  }

  $split_stats_result->save_stats_xml(qq{$qc_output_dir/$both_fastq}.qq{_$ref_name}.q{_split_stats.xml});
  $split_stats_result->store(qq{$qc_output_dir/$both_fastq}.qq{_$ref_name}.q{_split_stats.json});

  #
  # generate srf filter file from two hashes
  #
  my ($filter_file_hit, $filter_file_non_hit);

  if($srf_filter){

    print {*STDERR} "Generating srf filter files in output directory\n" or croak $ERRNO;
    $filter_file_hit = $output_dir;
    $filter_file_non_hit = $output_dir;

  }elsif($srf_filename){

    print {*STDERR} "Generating srf filter files in temp directory\n" or croak $ERRNO;
    $filter_file_hit = $work_dir;
    $filter_file_non_hit = $work_dir;
  }

  if($srf_filter || $srf_filename){

    $filter_file_hit .= q{/}.$both_fastq.qq{_$ref_name.filter};
    $filter_file_non_hit .=q{/}.$both_fastq.qq{_non$ref_name.filter};

    $self->_srf_filter_file($hit_sequence_ids1, $filter_file_hit);
    $self->_srf_filter_file($nonhit_sequence_ids1, $filter_file_non_hit);

  }

  #
  # split srf file if need
  #
  if($srf_filename && !$stats_only){

    my $args_srf_split = {
                srf_filename        => $srf_filename,
                filter_file_hit     => $filter_file_hit,
                filter_file_non_hit => $filter_file_non_hit,
                output_dir          => $output_dir,
               };
    $self->split_srf_file($args_srf_split);
  }

  return 1;
}

sub get_filename_no_extension{
  my ($self, $original_filename) = @_;

  my $filename = basename($original_filename);
  $filename =~ s/[.][^.]*$//mxs;
  return $filename;
}

sub split_srf_file{
  my ($self, $args) = @_;
  my $srf_filename = $args->{srf_filename};
  my $filter_file_hit = $args->{filter_file_hit};
  my $filter_file_non_hit = $args->{filter_file_non_hit};
  my $output_dir = $args->{output_dir};
  my $ref_name = $args->{ref_name};

  if(!$ref_name){
    $ref_name = q{human};
  }
  print {*STDERR} "Spliting srf file\n" or croak $ERRNO;
  my $srf_human = $output_dir.q{/}.$self->get_filename_no_extension($srf_filename).qq{_$ref_name.srf};
  my $srf_non_human = $output_dir.q{/}.$self->get_filename_no_extension($srf_filename).qq{_non$ref_name.srf};

  $self->srf_filter($srf_filename, $filter_file_hit, $srf_human );
  $self->srf_filter($srf_filename, $filter_file_non_hit, $srf_non_human);

  return 1;
}

sub bwa_alignment_list_fork{
  my ($self, $args_ref) = @_;

  my $fastq_list = $args_ref->{fastq_list};

  my $count = 0;
  my $pm = Parallel::ForkManager->new(scalar @{$fastq_list});

  foreach my $fastq (@{$fastq_list}){

    $count++;
    $pm->start and next; # do the fork

    if($fastq){
      $self->bwa_alignment($args_ref, $count);
    }

    $pm->finish; # do the exit in the child process
  }

  $pm->wait_all_children;

  return 1;
}

sub bwa_alignment_list{
  my ($self, $args_ref) = @_;

  my $fastq_list = $args_ref->{fastq_list};

  my $count = 1;

  foreach my $fastq (@{$fastq_list}){

    if($fastq){

      $self->bwa_alignment($args_ref, $count);
    }

    $count++;
  }

  return 1;
}

sub bwa_alignment{
  my ($self, $args_ref, $count) = @_ ;

  my $fastq = $args_ref->{fastq_list}->[$count - 1];
  my $human_ref_bwa = $args_ref->{human_ref_bwa};

  my $bam_out   = $args_ref->{q{bam_outputfile}.$count};

  my $hit_sequence_ids = {};
  my $nonhit_sequence_ids = {};
  tie %{ $hit_sequence_ids }, 'DB_File', $args_ref->{q{hit_filename}.$count} or croak "Cannot tie hash to a file\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)
  tie %{ $nonhit_sequence_ids }, 'DB_File', $args_ref->{q{nonhit_filename}.$count} or croak "Cannot hash to a file\n";## no critic (Perl::Critic::Policy::Miscellanea::ProhibitTies)

  my $alignment = npg_common::Alignment->new({
                                bwa_cmd     => $self->bwa_cmd(),
                                bwa_options => $self->bwa_options(),
                                samtools_cmd => $self->samtools_cmd(),
                                });

  my $args_ref2 = {ref_root => $human_ref_bwa,
                   fastq    => $fastq,
                   bam_out  => $bam_out,
                   hit_sequence_ids => $hit_sequence_ids,
                   nonhit_sequence_ids => $nonhit_sequence_ids,
                 };

  $alignment->bwa_align_se($args_ref2);

  my $sorted_bam = $bam_out.q{.sorted};
  $alignment->sort_bam($bam_out, $sorted_bam);
  my $move_rs = system "mv $sorted_bam.bam $bam_out";
  if($move_rs != 0){
     croak "mv $sorted_bam.bam $bam_out failed";
  }
  return 1;
}

sub coverage_depth{
  my ($self, $bam_out, $human_ref_fa) = @_;

  #transfer sam to sorted bam and do pile up to get stats
  my $alignment = npg_common::Alignment->new({
                                samtools_cmd =>$self->samtools_cmd(),
                                });

  my ($chromosome_coverage, $chromosome_depth) = $alignment->alignment_coverage_from_sam($bam_out, $human_ref_fa);

  return ($chromosome_coverage, $chromosome_depth);
}

sub fastq_split {
  my ($self, $fastq_fn, $select_ids, $select_fn, $rest_fn) = @_;

  $fastq_fn = $self->_detaint_filename($fastq_fn);
  $select_fn = $self->_detaint_filename($select_fn);
  $rest_fn = $self->_detaint_filename($rest_fn);

  my $select_fastqcheck = $select_fn.q{check};
  my $rest_fastqcheck   = $rest_fn.q{check};

  my $select_md5 = $select_fn.q{.md5};
  my $rest_md5   = $rest_fn.q{.md5};

  ## no critic (RequireBriefOpen ProhibitTwoArgOpen )
  open my $fastq_fh, q[<], $fastq_fn or croak "$fastq_fn:\n$ERRNO";

  open my $select_fh, q[>], $select_fn or croak "$select_fn:\n$ERRNO";
  open my $rest_fh, q[>], $rest_fn or croak "$rest_fn:\n$ERRNO";

  open my $select_check_fh, qq{| fastqcheck > $select_fastqcheck} or croak "Failed to pipe to $select_fastqcheck:\n$ERRNO";
  open my $rest_check_fh, qq{| fastqcheck > $rest_fastqcheck} or croak "Failed to pipe to $rest_fastqcheck:\n$ERRNO";

  open my $select_md5_fh, qq{| md5sum > $select_md5} or croak "Failed to pipe to $select_md5:\n$ERRNO";
  open my $rest_md5_fh, qq{| md5sum > $rest_md5} or croak "Failed to pipe to $rest_md5:\n$ERRNO";
  ## use critic

  my $select_dest = IO::Tee->new( $select_fh, $select_check_fh, $select_md5_fh);
  my $rest_dest   = IO::Tee->new( $rest_fh, $rest_check_fh,$rest_md5_fh );

  while(my $id = <$fastq_fh>) {

    chomp $id;
    my $id_original = $id;
    ($id) = $id =~ /^@([\w:\-\#]+)\/?\w?$/mxs;

    my $seq = <$fastq_fh>;
    my $seq_name  = <$fastq_fh>;
    my $qual = <$fastq_fh>;

    if($select_ids->{$id}){

      print {$select_dest} "$id_original\n$seq$seq_name$qual" or croak $ERRNO;
    }else{

      print {$rest_dest} "$id_original\n$seq$seq_name$qual"or croak $ERRNO;
    }

  }

  close $fastq_fh or croak $ERRNO;
  $select_dest->close() or croak $ERRNO;
  $rest_dest->close() or croak $ERRNO;

  return 1;
}

sub srf_filter{
  my ($self, $srf_fn, $filter_fn, $out_fn) = @_;

  $srf_fn = $self->_detaint_filename($srf_fn);
  $filter_fn= $self->_detaint_filename($filter_fn);
  $out_fn = $self->_detaint_filename($out_fn);

  my $srf_filter_cmd = $self->srf_filter_cmd().qq{ -f file=$filter_fn $srf_fn $out_fn};
  print {*STDERR} ">$srf_filter_cmd\n" or croak "$ERRNO";

  my $srf_filter_cmd_rs = system $srf_filter_cmd;

  if($srf_filter_cmd_rs != 0){

    croak "srf_filter failed for $srf_fn\n";
  }

  return $srf_filter_cmd_rs;
}

sub _srf_filter_file{
  my ($self, $ids_hash, $output_fn) = @_;

  $output_fn = $self->_detaint_filename($output_fn);
  open my $output_fh, q[>], $output_fn or croak "$output_fn:\n$ERRNO";

  while( my ($id, $value) = each %{$ids_hash} ){
    print {$output_fh} "read=$id\n" or croak 'cannot print';
  }

  close $output_fh or  croak $ERRNO;

  return 1;
}

sub hash_size{
  my ($self, $hash) = @_;

  my $count = 0;
  while( my ($id, $value) = each %{$hash} ){
    $count++;
  }
  return $count;
}

sub _get_seq_ids_from_file {
  my ($self, $id_fn) = @_;

  my %seq_ids = ();

  open my $id_fh, q[<], $id_fn or croak "$id_fn:\n$ERRNO";

  while(my $id = <$id_fh>) {

    chomp $id;
    $seq_ids{$id}++;
  }
  close $id_fh or croak $ERRNO;

  return \%seq_ids;
}

sub _detaint_filename {

  my ($self, $string) = @_;

  my ($dt) = $string =~ m{^ ( [#[:word:]./\\+-]+ ) $}msx;

  if ($dt ne $string) {

    croak "could not detaint $string, got $dt";
  }

  return $dt;
}

1;

__END__

=head1 NAME

npg_common::Fastq_Split

=head1 VERSION

$Revision: 13877 $

=head1 SYNOPSIS

  my $oFastqSplit = npg_common::Fastq_Split->new();

=head1 DESCRIPTION

Class to split fastq files or srf file given a list of ids

=head1 SUBROUTINES/METHODS

=head2 new - constructor to create a object

=head2 coverage_depth - given a bam file calulate alignment coverage and depth

=head2 file_split - given one or two fastq files, or srf file, split them into human and nonhuman files

=head2 fastq_split - given a fastq file and a list ids in hash, and two output file names, split the files into two

=head2 bwa_alignment - given a fastq file, do the bwa alignment

=head2 bwa_alignment_list - given a list of fastq files, do the bwa alignment for each of them

=head2 bwa_alignment_list_fork - given a list of fastq files, fork out to do the bwa alignment for each of them

=head2 srf_filter - given a list of ids in a file, call srf_filter command to generate a new srf file based on the id list

=head2 split_srf_file - given a srf file, two sets of ids and output_dir, using srf_filter command to split the srf file into two

=head2 srf_filter_file - given a list of ids in hash, generate srf_filter file

=head2 hash_size - return size of a given hash ref, instead of calling keys just in case the tied hash is too big

=head2 _detaint_filename - check file name

=head2 get_filename_no_extension - remove file's path and extension name

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

strict
warnings
English
Carp
Class::Accessor
npg_common::Alignment
npg_common::Hit_Sequence_SAM
File::Basename
File::Temp
Readonly
DB_File
Parallel::ForkManager

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi, E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 GRL, by Guoying Qi

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.

=cut
