#########
# Author:        Marina Gourtovaia
# Maintainer:    $Author: ajb $
# Created:       June 2010
# Last Modified: $Date: 2011-08-10 10:05:48 +0100 (Wed, 10 Aug 2011) $
# Id:            $Id: map2lims.pm 13877 2011-08-10 09:05:48Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-38.0/lib/npg_common/roles/run/lane/map2lims.pm $
#

package npg_common::roles::run::lane::map2lims;

use strict;
use warnings;
use Moose::Role;
use Carp;
use English qw(-no_match_vars);
use Readonly;

use npg::api::run_lane;
use st::api::lane;
use st::api::asset;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13877 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd ProhibitParensWithBuiltins RequireCheckingReturnValueOfEval)

=head1 NAME

npg_common::roles::run::lane::map2lims

=head1 VERSION

$Revision: 13877 $

=head1 SYNOPSIS

=head1 DESCRIPTION

Interface (Moose role) that, given information about a lane,
retrieves LIMS object that are needed for specific reasons,
for example, for finding a reference.
Requires the following method present in the implementing class: id_run, position.

=head1 SUBROUTINES/METHODS

=cut


=head2 asset

Asset object

=cut
has 'asset'  =>   (isa             => 'Maybe[st::api::asset]',
                   is              => 'ro',
                   required        => 0,
                   writer          => '_set_asset',
                   lazy_build      => 1,
                  );
sub _build_asset {
    my $self = shift;
    $self->_build_asset_and_samples;
    return $self->asset;
}

=head2 samples

Sample objects

=cut
has 'samples'  => (isa             => 'ArrayRef',
                   is              => 'ro',
                   required        => 0,
                   writer          => '_set_samples',
                   lazy_build      => 1,
                  );
sub _build_samples {
    my $self = shift;
    $self->_build_asset_and_samples();
    return $self->samples;
}


=head2 st_lane

st::api::lane onject

=cut
has 'st_lane'  => (isa             => 'st::api::lane',
                   is              => 'ro',
                   required        => 0,
                   lazy_build      => 1,
                  );
sub _build_st_lane {
  my $self = shift;

  $self->_check_prerequisites();
  my $hash_ref = {id_run => $self->id_run, position => $self->position,};
  if ( $self->can( q{has_api_util} ) && $self->has_api_util() ) {
    $hash_ref->{util} = $self->api_util();
  }
  my $run_lane = npg::api::run_lane->new( $hash_ref );

  if (!$run_lane) {
    croak q[run_lane object undefined for run ] . $self->id_run . q[, lane ] . $self->position;
  }

  my $lane = $run_lane->st_lane;
  if ($lane->pool && !$lane->pool_description) {
    croak q[No pool description for run ] . $self->id_run . q[, lane ] . $self->position;
  }

  return $lane;
}


sub _check_prerequisites {
  my $self = shift;
  eval {
    $self->id_run;
    $self->position;
    1;
  } or do {
    croak q[id_run and position attributes should be defined for the object that uses functions from npg_common::roles::run::lane::map2lims ];
  };
  return;
}

sub _build_asset_and_samples {
  my $self = shift;

  my $tag_index = undef;
  eval { $tag_index = $self->tag_index; };

  my $lane = $self->st_lane;
  if ($lane->pool && $tag_index) {
    if (!$lane->pool_description) {
      croak q[No pool description for run ] . $self->id_run . q[, lane ] . $self->position;
    }
    my $error_end =  qq[tag $tag_index in run ] . $self->id_run . q[, lane ] . $self->position;

    if (!exists $lane->pool_description->{$tag_index}) {
      croak qq[No plex description for $error_end];
    }
    my $plex = $lane->pool_description->{$tag_index};

    my $lib_tube_asset_id = $plex->{lib_tube_asset_id};
    if (!$lib_tube_asset_id) {
      croak qq[No asset id in $error_end];
    }
    my $hash_st_ref = {'id' => $lib_tube_asset_id,};
    if ( $self->can( q{has_api_util} ) && $self->has_api_util() ) {
      $hash_st_ref->{util} = $self->api_util();
    }
    $self->_set_asset(st::api::asset->new( $hash_st_ref ));

    if (exists $plex->{sample} && $plex->{sample}) {
      $self->_set_samples([$plex->{sample}]);
    } else {
      $self->_set_samples([]);
    }
  } else {
    $self->_set_samples($lane->samples);
    $self->_set_asset($lane->entity);
  }
  return;
}

=head2 asset4expected_insert_size

LIMS asset object (LibraryTube) for finding expected insert size

=cut
sub asset4expected_insert_size {
  my $self = shift;
  return $self->asset;
}


=head2 assets4ref

A reference to a list of LIMS assets that are needed to find a genome reference
for a lane or, if a tag_index attribute is defined, for an indexed component
of a pool.

given a parameter 'library', this method will return a list
only including the corresponding library object st::api::asset

=cut
sub assets4ref {
  my ($self, $library_sample) = @_;
  return !$library_sample || $library_sample eq q[sample] ?  $self->samples : [$self->asset];
}

no Moose::Role;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item strict

=item warnings

=item Moose::Role

=item Carp

=item English

=item Readonly

=item npg::api::run_lane

=item st::api::lane

=item st::api::asset

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Author: Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
