#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2010 06 21
# Last Modified: $Date: 2011-08-10 10:05:48 +0100 (Wed, 10 Aug 2011) $
# Id:            $Id: BAM_MarkDuplicate.pm 13877 2011-08-10 09:05:48Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-38.0/lib/npg_common/sequence/BAM_MarkDuplicate.pm $
#

package npg_common::sequence::BAM_MarkDuplicate;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Temp qw/ tempfile tempdir /;
use Perl6::Slurp;
use IPC::Open3;
use npg_qc::autoqc::results::bam_flagstats;
use npg_qc::autoqc::checks::check;
use Cwd qw(abs_path);
use File::Basename;

with 'MooseX::Getopt';

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13877 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $DEFAULT_COMMAND_OPTIONS => {
                               VALIDATION_STRINGENCY => 'SILENT',
                               VERBOSITY             => 'ERROR',
                               ASSUME_SORTED         => 'true',
                               READ_NAME_REGEX       => '[a-zA-Z0-9_]+:[0-9]:([0-9]+):([0-9]+):([0-9]+).*',
                               REMOVE_DUPLICATES     => 'false',
                               CREATE_MD5_FILE       => 'TRUE',
                               CREATE_INDEX          => 'TRUE',
                               MAX_FILE_HANDLES_FOR_READ_ENDS_MAP => 900,
                                                 };

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::BAM_MarkDuplicate

=head1 VERSION

$LastChangedRevision: 13877 $

=head1 SYNOPSIS

  my $bam = npg_common::sequence::BAM_MarkDuplicate->new({
                 input_bam     => 'input.bam',
                 output_bam    => 'output.bam',
                 metrics_json  => 'metrics.json',

               });
  $bam->process();

=head1 DESCRIPTION


=head1 SUBROUTINES/METHODS

=cut

=head2 java_command

JAVA command

=cut
has 'java_cmd'       => ( isa             => 'Str',
                          is              => 'rw',
                          default         => '/software/bin/java',
                          documentation   => 'java command',
                        );

=head2 java_Xmx

JAVA command Xmx value

=cut
has 'java_Xmx'       => ( isa             => 'Int',
                          is              => 'rw',
                          default         => '3000',
                          documentation   => 'java command Xmx value',
                        );

=head2 jar_file

Picard MarkDuplicates jar file

=cut

has 'jar_file'       => ( isa             => 'Str',
                          is              => 'rw',
                          default         => abs_path ('/software/solexa/bin/aligners/picard/current/MarkDuplicates.jar'),
                          trigger        => \&_covert_picard_path,
                          documentation   => 'Picard MarkDuplicates jar file',
                        );

sub _covert_picard_path {
   my ($self) = @_;

   my $old_path =  $self->jar_file();
   my $abs_path = abs_path($old_path);

   if($abs_path ne $old_path){
      $self->jar_file($abs_path);
   }
   return;
}

=head2 mark_duplicate_options

=cut

has 'mark_duplicate_options'       => ( isa             => 'HashRef',
                                        is              => 'rw',
                                        default         => sub { $DEFAULT_COMMAND_OPTIONS },
                                        documentation   => 'Picard MarkDuplicates command options',
                                      );
=head2 input_bam

input bam file name with path

=cut

has 'input_bam'      => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'input bam filename with path',
                        );

=head2 output_bam

output bam with path

=cut

has 'output_bam'     => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'output bam filename with path',
                        );
=head2 metrics_file

output metrics file name

=cut

has 'metrics_file'   => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 0,
                          lazy_build      => 1,
                          documentation   => 'output metrics file name',
                        );

sub _build_metrics_file {

   my ($fh, $filename) = tempfile(UNLINK => 1);
   return $filename;
}

=head2 temp_dir

temp dir

=cut

has 'temp_dir'       => ( isa             => 'Str',
                          is              => 'rw',
                          lazy_build      => 1,
                          documentation   => 'temp dir',
                        );
sub _build_temp_dir {
   my $self = shift;
   #return tempdir( CLEANUP => 1 );
   my $output = $self->output_bam();
   my ($name,$path,$suffix) = fileparse($output);
   return tempdir(CLEANUP => 1, DIR => $path);
}

=head2 mark_duplicate_cmd

construct the whole picard markduplicate command

=cut
has 'mark_duplicate_cmd'  => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'read only mark duplicate command',
                             );

sub _build_mark_duplicate_cmd {
   my $self = shift;

   my $cmd = $self->java_cmd.q{ -Xmx}.$self->java_Xmx().q{m};
   $cmd .= q{ -jar }.$self->jar_file();
   $cmd .= q{ INPUT=}.$self->input_bam();
   $cmd .= q{ OUTPUT=}.$self->output_bam();
   $cmd .= q{ METRICS_FILE=}.$self->metrics_file();
   $cmd .= q{ TMP_DIR=}.$self->temp_dir();

   my $options = $self->mark_duplicate_options();
   foreach my $option_key (keys %{$options}){
       $cmd .= qq{ $option_key='}.$options->{$option_key}.q{'};
   }
   return $cmd;
}

=head2 samtools_cmd

samtools command

=cut
has 'samtools_cmd'        => ( isa             => 'Str',
                               is              => 'ro',
                               default         => abs_path('/software/solexa/bin/aligners/samtools/current/samtools'),
                               documentation   => 'samtools command',
                             );
=head2 result

autoqc result object

=cut
has 'result'              => ( isa             => 'npg_qc::autoqc::results::bam_flagstats',
                               is              => 'rw',
                               lazy_build      => 1,
                               documentation   => 'autoqc result object',
                             );
sub _build_result {
   my $self = shift;
   my $result= npg_qc::autoqc::results::bam_flagstats->new();
   if($self->id_run()){
      $result->id_run($self->id_run);
   }
   if($self->position()){
      $result->position($self->position);
   }
   if(defined $self->tag_index){
      $result->tag_index($self->tag_index);
   }
   if($self->human_split){
      $result->human_split($self->human_split);
   }
   return $result;
}

=head2 metrics_json

metrics json output file name

=cut
has 'metrics_json'   => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'output json metrics file name',
                        );
=head2 position

Lane number. An integer from 1 to 8 inclusive.

=cut
has 'position'    => (isa       => 'Int',
                      is        => 'rw',
                      required  => 0,
                     );


=head2 id_run

Run id for the lane to be checked.

=cut
has 'id_run'      => (
                       isa            => 'Int',
                       is             => 'rw',
                       required       => 0,
		      );


=head2 tag_index

Tag index. An integer from 0 to 10000. Zero means that no tag has been matched.

=cut
has 'tag_index'   => (isa        => 'Int',
                      is         => 'rw',
                      required   => 0,
                     );
=head2 human_split

to show the bam file is human part or nonhuman
default 'all', which means not split

=cut
has 'human_split'   => (isa        => 'Str',
                        is         => 'rw',
                        required   => 0,
                       );

=head2 replace_file

replace the input bam file with the output marked bam file

=cut
has 'replace_file'=> (isa        => 'Bool',
                      is         => 'rw',
                      required   => 0,
                     );

=head2 process

main method to call
=cut

sub process {
  my $self = shift;

  if( ! -e $self->input_bam() ){
     croak 'Input bam file does not exist to mark duplicate: '.$self->input_bam();
  }

  my $mark_duplicate_cmd = $self->mark_duplicate_cmd();
  print {*STDERR} "$mark_duplicate_cmd\n" or croak $ERRNO;

  my $mark_duplicate_rs = system $mark_duplicate_cmd;
  if( $mark_duplicate_rs != 0){
    croak "Picard MarkDuplicates failed:\n$mark_duplicate_cmd";
  }

  $self->result->parsing_metrics_file($self->metrics_file());

  $self->_flagstats();
  $self->_version_info();

  $self->result->store($self->metrics_json);

  if($self->replace_file){

    $self->_move_file($self->output_bam, $self->input_bam);

    my $md5_file_name_mk = $self->output_bam.q{.md5};
    my $md5_file_name     = $self->input_bam.q{.md5};

    if( -e $md5_file_name_mk ) {
       $self->_move_file($md5_file_name_mk, $md5_file_name);
    }else{
       print {*STDERR}  "BAM MD5 file not generated: $md5_file_name\n" or croak $ERRNO;
    }

    my $index_file_name_mk = $self->output_bam;
    my $index_file_name    = $self->input_bam;
    $index_file_name_mk =~ s/bam$/bai/mxs;
    $index_file_name =~ s/bam$/bai/mxs;

    if( -e $index_file_name_mk ) {
       $self->_move_file($index_file_name_mk, $index_file_name);
    }else{
       print {*STDERR}  "BAM index file not generated: $index_file_name\n" or croak $ERRNO;
    }
  }

  return 1;
}

sub _move_file {
   my ($self, $source, $des) = @_;

   my $mv_cmd = q{mv }.$source.q{ }.$des;
   print {*STDERR} "$mv_cmd\n" or croak $ERRNO;
   my $mv_cmd_rs = system $mv_cmd;
   if ( $mv_cmd_rs != 0 ){
      croak "Failed: $mv_cmd";
   }
   return;
}

sub _flagstats {
  my $self = shift;

  my $samtools_cmd = $self->samtools_cmd().q{ flagstat }.$self->output_bam();

  print {*STDERR} "$samtools_cmd \n" or croak $ERRNO;

  my $pid = open3( undef, my $samtools_output, undef, $samtools_cmd );

  $self->result->parsing_flagstats($samtools_output);

  close $samtools_output or carp "Failed to close $samtools_output: $ERRNO";

  waitpid $pid, 0;

  return 1;
}

sub _version_info {
  my $self = shift;

  eval {
    my $samtools_version = npg_qc::autoqc::checks::check->aligner_version($self->samtools_cmd());
    if($samtools_version){
       $self->result->set_info('Samtools', $samtools_version);
    }
    1;
  } or do {
    carp $EVAL_ERROR;
  };

  my $picard_path = abs_path($self->jar_file());
  my ($picard_version) = $picard_path =~ m/ picard-tools- ( \d+ [.] \d+ )/mxs;
  if($picard_version){
     $self->result->set_info('Picard-tools', $picard_version);
  }

  return 1;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item Perl6::Slurp;

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
